#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>

#define NAM	100 //nombre max d'aretes sortant d'un sommet
#define NSM	1000 //nombre max de sommets

#define NVS	20 //nombre max de fois que l'on passe par un sommet
#define PASSING	true
#define BOUCLE true //autorise les retours intermédiaire à l'état de départ

#define NL 24 //nombre de lignes par page affichées

#define NO25 true //pas de passes en 2 ni 5

int max (int a, int b)
{
	if (a < b)
		return b;
	return a;
}

struct Sommet
{
	int a[NAM]; //aretes
	int n; //nombre d'aretes
	double l[NAM]; //labels
};

struct Graph
{
	Sommet s[NSM];
	int n; //nombre de sommets
	
	void Load (FILE *f);
	void List (FILE *d, int l); //écrit la liste des cycles non triviaux de longueur l
	void ListRec (FILE *d, int l, int i, int j, int r); //utilisé par List
	void ListSync (FILE *d, int l); //écrit la liste des cycles non triviaux de longueur l
	void ListSyncRec (FILE *d, int l, int i, int j, int r); //utilisé par List
	void Coups (int i); //écrit une plus petite suite de coups pour rejoindre l'état i depuis 0
	bool CoupsRec (int i, int j, int n);
};

#define EPSILON	.0001
void print (FILE *f, double d)
{
	if (d - (int)d < EPSILON)
		fprintf(f, "%d", (int)d);
	else if (2*d - (int)(2*d) < EPSILON)
		fprintf(f, "%.1lf", d);
	else
		fprintf(f, "%.2lf", d);
}

void printSync (FILE *f, double d, int p)
{
	if (d - (int)d < EPSILON)
		fprintf(f, "%d", (int)d);
	else if (2*d - (int)(2*d) < EPSILON)
		fprintf(f, "%dp", (int)(d+p-.5));
	else
		fprintf(f, "err");
}

double start[100]; //liste des coups à faire pour rejoindre l'état excité
int nstart = 0;
bool Graph::CoupsRec (int i, int j, int n)
{
	if (i == j)
		return true;
	if (n == 0)
		return false;
	int k;
	for (k=0;k<s[j].n;k++)
	{
		start[nstart] = s[j].l[k];
		nstart++;
		if (CoupsRec (i, s[j].a[k], n-1))
		{
			//print(d, s[j].l[k]);
			//fprintf(d, "	");
			//fprintf(d, "%lf	", s[j].l[k]);
			return true;
		}
		nstart--;
	}
	return false;
}

//écrit une plus petite suite de coups pour rejoindre l'état i depuis 0
void Graph::Coups (int i)
{
	nstart = 0;
	int n;
	for (n=0;n<100;n++)
	{
		if (CoupsRec(i, 0, n))
			break;
	}
}

void Graph::Load (FILE *f)
{
	n = 0;
	int i,j;
	double d;
	for (i=0;i<NSM;i++)
	{
		s[i].n = 0;
	}
	char tamp[1024];
	char *ptr;
	do
	{
		fgets(tamp, 1024, f);
		ptr = tamp;
		while (*ptr == ' ' || *ptr == '	')
			ptr++;
		if (sscanf(ptr, "%d -> %d [label=\"%lf", &i, &j, &d) == 3)
		{
			printf("%d -> %d\n", i, j);
			s[i].a[s[i].n] = j;
			s[i].l[s[i].n] = d;
			s[i].n++;
			n = max(i+1, n);
		}
	}while(!feof(f));
}

bool seen[NSM];
int seen2[NSM];
double lab[NAM];

int nstartmax = 0;
int lmax = 0;
int di = 0;

void Start (FILE *d)
{
	int k;
	fprintf(d, "	\\begin{tabular}{|");
	for (k=0;k<nstartmax;k++)
	{
		fprintf(d, "c");
	}
	for (k=0;k<lmax*(1+(lmax%2));k++)
	{
		//if (k == nstart)
		//	fprintf(d, "|");
		fprintf(d, "c");
	}
	fprintf(d, "|}\n		\\hline\n");
}

void StartSync (FILE *d)
{
	int k;
	fprintf(d, "	\\begin{tabular}{|");
	for (k=0;k<nstartmax;k++)
	{
		fprintf(d, "c");
	}
	for (k=0;k<lmax;k++)
	{
		//if (k == nstart)
		//	fprintf(d, "|");
		fprintf(d, "c");
	}
	fprintf(d, "|}\n		\\hline\n");
}

void Disp (FILE *d, int l) //affiche le pattern
{
	nstartmax = max(nstart, nstartmax);
	lmax = max(l*(1+l%2), lmax);
	
	di++;
	
	if (!d)
		return;
	int k, k0, m, li;
	for (li=0;li<2;li++) //en deux lignes
	{
		//écrit l'éventuel démarrage
		fprintf(d, "		");
		for (k=0;k<nstartmax;k++)
		{
			if (k == nstartmax-nstart && k == 0 && nstartmax > 1)
			{
				fprintf(d, "\\multicolumn{1}{|c}{");
			}else if (k == nstartmax-nstart && nstartmax == 1 && nstart == 1)
			{
				fprintf(d, "\\multicolumn{1}{|c|}{");
			}else if (k == nstartmax-1 && nstart >= 1 && nstartmax > 1)
			{
				fprintf(d, "\\multicolumn{1}{c|}{");
			}
			if (k >= nstartmax-nstart)
			{
				k0 = k - (nstartmax-nstart);
				if (k0%2 == li)
				{
					fprintf(d, "$");
					print(d, start[k0]);
					fprintf(d, "$");
				}
			}
			if (((k == nstartmax-nstart && k == 0) || k == nstartmax-1) && nstart >= 1)
				fprintf(d, "}");
			fprintf(d, "	& ");
		}
		//écrit le pattern en le dupliquant si nécessaire
		for (m=0;m<1+l%2;m++)
		{
			for (k=0;k<l;k++)
			{
				if ((k+m*l+nstart)%2 == li)
				{
					fprintf(d, "$");
					print(d, lab[k]);
					fprintf(d, "$");
				}
				if (k+l*m != lmax-1)
					fprintf(d, "	& ");
			}
		}
		//termine la ligne
		for (k=l*(1+l%2);k<lmax;k++)
		{
			if (k != lmax-1)
				fprintf(d, "	& ");
		}
		fprintf(d, " \\\\\n");
	}
	fprintf(d, "		\\hline\n");
}

void DispSync (FILE *d, int l) //affiche le pattern synchrone
{
	//réparti les passes entre les deux passeurs
	double p[2][NSM];
	int st = 0;
	int lm;
	int i;
	for (i=0;i<nstart;i++)
	{
		p[i%2][i/2] = start[i];
	}
	if (nstart%2 != 0)
	{
		p[1][nstart/2] = lab[0];
		st = nstart/2+1;
	}else
		st = nstart/2;
	for (i=0; i<l*(1+l%2); i++)
	{
		p[i%2][st + i/2] = lab[(i + nstart%2)%l];
	}
	lm = l/(1+(l+1)%2);
	
#if NO25
	//détermine si le pattern doit être affiché
	for (i=0;i<st+lm;i++)
	{
		if ((p[0][i] - (int)p[0][i] >= EPSILON && p[0][i]-.5 <= 2) || (p[1][i] - (int)p[1][i] >= EPSILON && p[1][i]+.5 >= 5))
			return;
	}
#endif

	di++;
	
	nstartmax = max(st, nstartmax);
	lmax = max(lm, lmax);
	if (!d)
		return;
	
	//affichage
	int li, k, k0, m;
	for (li=0;li<2;li++) //en deux lignes
	{
		//écrit l'éventuel démarrage
		fprintf(d, "		");
		for (k=0;k<nstartmax;k++)
		{
			if (k == 0)
			{
				fprintf(d, "\\multicolumn{%d}{|", nstartmax);
				for (m=0;m<nstartmax;m++)
				{
					fprintf(d, "c");
				}
				fprintf(d, "|}{");
			}
			if (k >= nstartmax-st)
			{
				k0 = k - (nstartmax-st);
				fprintf(d, "$");
				printSync(d, p[li][k0], li);
				fprintf(d, "$");
			}
			if (k == nstartmax-1)
				fprintf(d, "}");
			fprintf(d, "	& ");
		}
		//écrit le pattern
		for (k=0;k<lm;k++)
		{
			fprintf(d, "$");
			printSync(d, p[li][k+st], li);
			fprintf(d, "$");
			if (k != lmax-1)
				fprintf(d, "	& ");
		}
		//termine la ligne
		for (k=lm;k<lmax;k++)
		{
			if (k != lmax-1)
				fprintf(d, "	& ");
		}
		fprintf(d, " \\\\\n");
	}
	fprintf(d, "		\\hline\n");
}

void End (FILE *d)
{
	fprintf(d, "	\\end{tabular}\n\n");
}

//i : sommet de départ
//j : sommet courant
//r : nombre d'étapes effectuées
//l : longueur des cycles recherchés
void Graph::ListRec (FILE *d, int l, int i, int j, int r)
{
	if (seen[i])
		return;
	if (!seen2[i])
		return;
	
#if !BOUCLE
	if (r == 1 && i == j)
		return; //on a pas le droit de faire une boucle de taille 1 dès le début
	if (i == j && l != r && r > 0)
		return;
#endif
	seen2[i]--;
	
	int k;
	if (l == r) //on a déjà un chemin de la longueur voulue
	{
		if (i == j) //on a bien un cycle
		{
			bool pass = true;
#if PASSING
			//vérifie que le cycle contient au moins une passe
			pass = false;
			for (k=0;k<l;k++)
			{
				if (lab[k] - (int)lab[k] >= EPSILON)
				{
					pass = true;
					break;
				}
			}	
#endif
			if (pass)
			{
				//cherche la plus petite suite de coups pour rejoindre l'état
				Coups(i);
				//affiche le cycle
				Disp(d, l);
				if (di%NL == NL-1)
				{
					if (d)
					{
						End(d);
						Start(d);
					}
				}
				/*
				for (k=0;k<l;k++)
				{
					print(d, lab[k]);
					fprintf(d, "	");
				}*/
				fflush(d);
			}
		}
	}else
	{
		//parcours les fils du sommet j
		for (k=0;k<s[j].n;k++)
		{
			lab[r] = s[j].l[k];
			ListRec (d, l, i, s[j].a[k], r+1);
		}
	}
	seen2[i]++;
}

//i : sommet de départ
//j : sommet courant
//r : nombre d'étapes effectuées
//l : longueur des cycles recherchés
void Graph::ListSyncRec (FILE *d, int l, int i, int j, int r)
{
	if (seen[j] || !seen2[i])
		return;
	if (r == 1 && i == j)
		return; //on a pas le droit de faire une boucle de taille 1 dès le début
#if !BOUCLE
	if (i == j && l != r && r > 0)
		return;
#endif
	seen2[i]--;
	
	int k;
	if (l == r) //on a déjà un chemin de la longueur voulue
	{
		if (i == j) //on a bien un cycle
		{
			bool pass = true;
#if PASSING
			//vérifie que le cycle contient au moins une passe
			pass = false;
			for (k=0;k<l;k++)
			{
				if (lab[k] - (int)lab[k] >= EPSILON)
				{
					pass = true;
					break;
				}
			}	
#endif
			if (pass)
			{
				//cherche la plus petite suite de coups pour rejoindre l'état
				Coups(i);
				//affiche le cycle
				DispSync(d, l);
				if (di%NL == NL-1)
				{
					if (d)
					{
						End(d);
						StartSync(d);
					}
				}
				//di++;
				/*
				for (k=0;k<l;k++)
				{
					print(d, lab[k]);
					fprintf(d, "	");
				}*/
				fflush(d);
			}
		}
	}else
	{
		//parcours les fils du sommet j
		for (k=0;k<s[j].n;k++)
		{
			lab[r] = s[j].l[k];
			ListSyncRec (d, l, i, s[j].a[k], r+1);
		}
	}
	seen2[i]++;
}

void Graph::List (FILE *d, int l)
{
	int i;
	for (i=0;i<NSM;i++)
	{
		seen[i] = false;
		seen2[i] = NVS;
	}
	for (i=0;i<NAM;i++)
	{
		lab[i] = 0;
	}
	//parcours la liste des sommets
	for (i=0;i<n;i++)
	{
		ListRec(d, l, i, i, 0);
		seen[i] = true;
	}
}

void Graph::ListSync (FILE *d, int l)
{
	int i;
	for (i=0;i<NSM;i++)
	{
		seen[i] = false;
		seen2[i] = NSM;
	}
	for (i=0;i<NAM;i++)
	{
		lab[i] = 0;
	}
	//parcours la liste des sommets
	for (i=0;i<n;i++)
	{
		ListSyncRec(d, l, i, i, 0);
		seen[i] = true;
	}
}

int main(int nargs, char **args)
{
	if (nargs < 4)
	{
		printf("Entrer le nombre du fichier dot, le nom du fichier résultat, puis la période max.\n");
		return 0;
	}
	
	FILE *f = fopen(args[1], "r");
	if (!f)
	{
		printf("Impossible d'ouvrir le fichier %s !\n", args[1]);
		return 0;
	}
	FILE *d = fopen(args[2], "w");
	if (!d)
	{
		printf("Impossible d'ouvrir le fichier %s !\n", args[2]);
		return 0;
	}
	
	Graph g;
	g.Load(f);
	
	int hmax = 0;
	if (sscanf(args[3], "%d", &hmax) != 1)
	{
		printf("Impossible de lire la période max : %s n'est pas un nombre.\n", args[3]);
		return 0;
	}
	
	//détermine les max
	int i;
	nstartmax = 0;
	lmax = 0;
	for (i=2;i<=hmax;i++)
	{
		g.List(NULL, i);
	}
	//fseek(d, 0, SEEK_SET);
	//écrit vraiment
	Start(d);
	di = 0;
	for (i=2;i<=hmax;i++)
	{
		g.List(d, i);
	}
	End(d);
	
	//détermine les max
	nstartmax = 0;
	lmax = 0;
	for (i=2;i<=hmax;i++)
	{
		g.ListSync(NULL, i);
	}
	//écrit vraiment
	StartSync(d);
	di = 0;
	for (i=2;i<=hmax;i++)
	{
		g.ListSync(d, i);
	}
	End(d);
	
	fclose(d);
	fclose(f);
	
	return 0;
}